import re
import pandas as pd
import numpy as np

def read_ldsc(filelist=[],mode="h2"):
#h2 mode
#####################################################################    
    if mode=="h2":
        summary = pd.DataFrame(columns = ['Filename', 'h2_obs', 'h2_se','Lambda_gc','Mean_chi2','Intercept','Intercept_se',"Ratio","Ratio_se"])
        
        for index, ldscfile in enumerate(filelist):
            print("Loading file "+str(index+1)+" :" + ldscfile +" ...")
            row={}
            
            with open(ldscfile,"r") as file:
                row["Filename"]=ldscfile.split("/")[-1]
                line=""
                while not re.compile('^Total Observed scale h2').match(line):
                    line = file.readline()
                    if not line: break
                        
                
                        ## first line h2 se
                objects = re.compile('[a-zA-Z\s\d]+:|[-0-9.]+|NA').findall(line)
                row["h2_obs"]=objects[1]
                row["h2_se"]=objects[2]

                ##next line lambda gc

                objects = re.compile('[a-zA-Z\s\d]+:|[-0-9.]+|NA').findall(file.readline())
                row["Lambda_gc"] = objects[1]
                ##next line Mean_chi2

                objects = re.compile('[a-zA-Z\s\d]+:|[-0-9.]+|NA').findall(file.readline())
                row["Mean_chi2"]=objects[1]
                ##next line Intercept

                objects = re.compile('[a-zA-Z\s\d]+:|[-0-9.]+|NA').findall(file.readline())
                row["Intercept"]=objects[2]
                row["Intercept_se"]=objects[2]
                ##next line Ratio
                
                lastline=file.readline()
                if re.compile('NA').findall(lastline):
                    row["Ratio"]="NA"
                    row["Ratio_se"]="NA"
                elif re.compile('<').findall(lastline):
                    row["Ratio"]="Ratio < 0"
                    row["Ratio_se"]="NA"
                else:
                    objects = re.compile('[a-zA-Z\s\d]+:|[-0-9.]+').findall(lastline)
                    row["Ratio"]=objects[1]
                    row["Ratio_se"]=objects[2]
            summary = summary.append(row,ignore_index=True)
###############################################################################
    if mode=="rg":
        summary = pd.DataFrame(columns = ['p1',
                                          'p2',
                                          'rg',
                                          'se' ,
                                          'z','p',
                                          'h2_obs','h2_obs_se',
                                          'h2_int','h2_int_se',
                                          'gcov_int','gcov_int_se']
                               )
        
        for index, ldscfile in enumerate(filelist):
            print("Loading file "+str(index+1)+" :" + ldscfile +" ...")
            
            with open(ldscfile,"r") as file:
                line=""
                while not re.compile('^Summary of Genetic Correlation Results').match(line):
                    line = file.readline()
                    if not line: break
                line = file.readline() # header        
                
                line = file.readline() #line1
                
                ## first line h2 se
                while line.strip():
                    row = line.split()
                    row_series = pd.Series(row, index = summary.columns)
                    summary = summary.append(row_series, ignore_index=True)
                    line = file.readline()
            
        summary[['rg','se' ,'z','p','h2_obs','h2_obs_se','h2_int','h2_int_se','gcov_int','gcov_int_se']]  = summary[['rg','se' ,'z','p','h2_obs','h2_obs_se','h2_int','h2_int_se','gcov_int','gcov_int_se']] .astype("float32")            
    return summary   


import pandas as pd
from os import path
def ldsc(insumstats,chrom=None,pos=None,rsid=None,verbose=True):
    if verbose:print("  - Processing "+str(len(insumstats))+" raw variants...")
    if rsid:
        sumstats = insumstats[rsid]
    elif chrom and pos:
        sumstats = insumstats.loc[:,[chrom,pos]].astype("string")
        sumstats["chr:pos"]=sumstats[chrom]+":"+sumstats[pos] 
    data_path = path.join(path.dirname(__file__), 'data/hapmap3_rs_chr_pos_a1_a2/hapmap3_db150_hg19.snplist.gz')
    hapmap3_ref = pd.read_csv(data_path,sep="\s+",dtype="string")
    
    if rsid:
        output=insumstats.loc[sumstats.isin(hapmap3_ref["rsid"].values),:]
        if verbose:print("  - Raw input contains "+str(len(output))+" hapmaps variants based on rsid...")
        return output
    elif chrom and pos:
        hapmap3_ref["chr:pos"]=hapmap3_ref["#CHROM"]+":"+hapmap3_ref["POS"]
        output=insumstats.loc[sumstats["chr:pos"].isin(hapmap3_ref["chr:pos"].values),:]
        if verbose:print("  - Raw input contains "+str(len(output))+" hapmaps variants based on chr:pos...")
        return output


import scipy.stats as ss
import numpy as np

def or2beta(insumstats,OR,OR_lower=None,OR_upper=None,OR_se=None,level=0.95):
    output = insumstats
    output["BETA"]=np.log(sumstats[OR])
    if OR_se:
        output["SE"]=np.log(sumstats[OR]+sumstats[OR_se])-np.log(sumstats[OR])
    elif OR_upper:
        tl,tu = ss.norm.interval(0.95,0)
        output["SE"]=(sumstats[upper] - sumstats[OR])/tu
    elif OR_lower:
        tl,tu = ss.norm.interval(0.95,0)
        output["SE"]=(sumstats[lower] - sumstats[OR])/tl
    return output
    



import numpy as np
from scipy.stats import norm
def h2_obs_to_liab(h2_obs, P, K, se_obs=None):
    '''
    Converts heritability on the observed scale in an ascertained sample to heritability
    on the liability scale in the population.
    Parameters
    ----------
    h2_obs : float
        Heritability on the observed scale in an ascertained sample.
    P : float in (0,1)
        Prevalence of the phenotype in the sample.
    K : float in (0,1)
        Prevalence of the phenotype in the population.
    se_obe:float
        se of h2_obs
    '''
    ## adopted from ldsc
    ## Reference:
    ## Estimating Missing Heritability for Disease from Genome-wide Association Studies
    ## https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3059431/
    
    if np.isnan(P) and np.isnan(K):
        return h2_obs
    if K <= 0 or K >= 1:
        raise ValueError('K must be in the range (0,1)')
    if P <= 0 or P >= 1:
        raise ValueError('P must be in the range (0,1)')
    
    # the fraction of y that is larger than t is K
    t = norm.isf(K)
    #z the height of the normal curve at point t
    z = norm.pdf(t)
    #Equation 23
    numerator= K**2 * (1-K)**2
    denominator= P*(1-P) * z**2
    conversion_factor = numerator/denominator
    
    if se_obs:
        se_lia = conversion_factor * se_obs
    if se_obs:
        return h2_obs * conversion_factor, se_lia
    return h2_obs * conversion_factor


import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import numpy as np
import scipy as sp
from gwaslab import getsig

def mqqplot(insumstats,
          chrom,
          pos,
          p,
          scaled=False,
          cut=0,
          cutfactor=10,
          cut_line_color="#ebebeb",
          windowsizekb=500,
          anno=None,
          sig_level=5e-8,
          sig_line_color="grey",
          suggestive_sig_level=5e-6,
          title =None,
          mtitle=None,
          qtitle=None,
          figsize =(15,5),
          fontsize = 10,
          colors = ["#000042", "#7878BA"],
          layout="qqm",
          verbose=True,
          repel_force=0.03,
          gc=True,
          save=None,
          saveargs={"dpi":300,"facecolor":"white"}        
          ):
    
# printing meta info ###############################
    if verbose: print("Basic settings:")
    if verbose: print("  - Genome-wide significance level is set to "+str(sig_level)+" ...")
    if verbose: print("  - Raw input contains "+str(len(insumstats))+" variants...")
    if verbose: print("Start conversion and QC:")
    
# read sumstats ###############################
    sumstats=pd.DataFrame()
    sumstats["P-value_association"]=insumstats[p].astype(np.float64)
    sumstats["CHROM"]=insumstats[chrom].astype(int)
    sumstats["POS"]=insumstats[pos].astype(int)
    if anno and anno!=True:
            sumstats["Annotation"]=insumstats[anno].astype(object)
            
# P value conversion ###############################  
    if scaled:
        if verbose:print("  - P values are already converted to -log10(P)!")
        sumstats["scaled_P"] = sumstats["P-value_association"]
    else:
        if verbose:print("  - P values are being converted to -log10(P)...")
        #sanity check
        outside_01_p_value_number=len(sumstats.loc[(sumstats["P-value_association"]>1)|(sumstats["P-value_association"]<=0),:])
        if verbose:print("  - Sanity check: "+ str(outside_01_p_value_number) +" variants with P value outside of (0,1] will be removed...")
        sumstats = sumstats.loc[(sumstats["P-value_association"]<=1)&(sumstats["P-value_association"]>0),:]
        sumstats["scaled_P"] = -np.log10(sumstats["P-value_association"])

    if verbose: print("  - Sanity check: "+str(len(sumstats[sumstats["scaled_P"].isin([np.nan, np.inf, -np.inf, float('inf'), None])])) + " na/inf/-inf variants will be removed..." )
    sumstats = sumstats[~sumstats["scaled_P"].isin([np.nan, np.inf, -np.inf, float('inf'), None])]

    # shrink at a certain value #############################################
    maxy = sumstats["scaled_P"].max()
    if verbose: print("  - Maximum -log10(P) values is "+str(maxy) +" .")
    if cut:
        if verbose: print("  - Minus log10(P) values above " + str(cut)+" will be shrunk with a shrinkage factor of " + str(cutfactor)+"...")
        maxticker=int(np.round(sumstats["scaled_P"].max()))
        sumstats.loc[sumstats["scaled_P"]>cut,"scaled_P"] = (sumstats.loc[sumstats["scaled_P"]>cut,"scaled_P"]-cut)/cutfactor +  cut
        maxy = (maxticker-cut)/cutfactor + cut

# create index ##########################

    if verbose:print("Plotting "+str(len(sumstats))+" variants:")
    #sort & add id
    sumstats = sumstats.sort_values(["CHROM","POS"])
    sumstats["id"]=range(len(sumstats))
    sumstats=sumstats.set_index("id")

    #create a position dictionary
    posdic=sumstats.groupby("CHROM")["POS"].max()
    posdiccul=dict(posdic)
    posdiccul[0]=0
    for i in range(2,sumstats["CHROM"].nunique()+1):
        posdiccul[i]=posdiccul[i-1]+posdiccul[i]

    #convert base pair postion to x axis position
    sumstats["add"]=sumstats["CHROM"].apply(lambda x : posdiccul[int(x)-1])
    sumstats["i"]=sumstats["POS"]+sumstats["add"]

    #for plot
    chrom_df=sumstats.groupby('CHROM')['i'].median()
    chrom_df=chrom_df+((chrom_df.index.astype(int))-1)*len(sumstats)*0.02

    sumstats["i"]=sumstats["i"]+((sumstats["CHROM"].astype(int))-1)*len(sumstats)*0.01

# assign marker size ##############################################
    sumstats["s"]=1
    sumstats.loc[sumstats["scaled_P"]>-np.log10(5e-4),"s"]=2
    sumstats.loc[sumstats["scaled_P"]>-np.log10(suggestive_sig_level),"s"]=3
    sumstats.loc[sumstats["scaled_P"]>-np.log10(sig_level),"s"]=10

# Plotting ########################################################  
    # manhattanplot : ax1
    if layout=="qqm":
        fig, (ax2, ax1) = plt.subplots(1, 2,gridspec_kw={'width_ratios': [1, 3]},figsize=figsize)
    elif layout=="mqq":
        fig, (ax1, ax2) = plt.subplots(1, 2,gridspec_kw={'width_ratios': [3, 1]},figsize=figsize)
        
    plot = sns.scatterplot(data=sumstats, x='i', y='scaled_P',
                       hue='CHROM',
                       palette=sns.color_palette(colors,n_colors=sumstats["CHROM"].nunique()),
                       legend=None,
                       size="s",
                       sizes=(10,40),
                       linewidth=0,
                       zorder=2,ax=ax1)   
        
    plot.set_xlabel('CHROM'); 
    plot.set_xticks(chrom_df);
    plot.set_xticklabels(chrom_df.index)
    
    sigline = plot.axhline(y=-np.log10(sig_level), linewidth = 2,linestyle="--",color=sig_line_color,zorder=1)
    if cut:
        cutline=plot.axhline(y=cut, linewidth = 2,linestyle="--",color=cut_line_color,zorder=1)
        plot.set_yticks([x for x in range(0,cut+1,2)]+[(maxticker-cut)/cutfactor + cut])
        plot.set_yticklabels([x for x in range(0,cut+1,2)]+[maxticker])

# Annotation #######################################################
    #!!!!!!!!!!!!!!! scale problem
    if anno and anno!=True:
        to_annotate=getsig(sumstats.loc[sumstats["scaled_P"]>-np.log10(sig_level)],
                           "Annotation",
                           'CHROM',
                           "POS",
                           "P-value_association",
                           sig_level=sig_level,
                           windowsizekb=windowsizekb,
                           verbose=False)
    else:
        to_annotate=getsig(sumstats.loc[sumstats["scaled_P"]>-np.log10(sig_level)],
                           "i",
                           'CHROM',
                           "POS",
                           "P-value_association",
                           windowsizekb=windowsizekb,
                           verbose=False,
                           sig_level=sig_level)
        
    if verbose:print("  - Found "+str(len(to_annotate))+" significant variants with a sliding window size of "+str(windowsizekb)+" kb...")

# Annotation #######################################################
    
    if anno and len(to_annotate)>0:
        #initiate a list for text and a starting position
        text = []
        last_pos=0
        
        for rowi,row in to_annotate.iterrows():
            # avoid text overlapping
            if row["i"]>last_pos+repel_force*sumstats["i"].max():
                last_pos=row["i"]
            else:
                last_pos+=repel_force*sumstats["i"].max()
            # data to pixels
            armB_length_in_point=plot.transData.transform((0, 0.95*maxy))[1]-plot.transData.transform((0, row["scaled_P"]+1))[1]
            
            #  
            armB_length_in_point= armB_length_in_point if armB_length_in_point>0 else plot.transData.transform((0, maxy+2))[1]-plot.transData.transform((0,  row["scaled_P"]+1))[1] 
            
            if anno==True:
                annotation_text=str(int(row["CHROM"])) +":"+ str(int(row["POS"]))
                annotation_col="CHR:POS"
            elif anno:
                annotation_text=row["Annotation"]
                annotation_col=anno
                
            text.append(plot.annotate(annotation_text,
                                         xy=(row["i"],row["scaled_P"]+1),
                                         xytext=(last_pos,1.15*maxy),rotation=40,
                                         ha="left",va="bottom",
                                         fontsize=fontsize,
                                         arrowprops=dict(arrowstyle="-|>",relpos=(0,0),color="#ebebeb",
                                         connectionstyle="arc,angleA=0,armA=0,angleB=90,armB="+str(armB_length_in_point)+",rad=0")
                                        )
                       )
        
        if verbose: print("  - Annotating using column "+annotation_col+"...")
    else:
        if verbose: print("  - Skip annotating")

    plot.set_ylabel("$-log_{10}(P)$",fontsize=fontsize)
    plot.set_xlabel("Chromosomes",fontsize=fontsize)
    plot.spines["top"].set_visible(False)
    plot.spines["right"].set_visible(False)
    plot.spines["left"].set_visible(True)
    
    if verbose: print("  - Created Manhattan plot successfully!")
    if mtitle and anno and len(to_annotate)>0: 
        pad=(plot.transData.transform((0, 1.35*maxy))[1]-plot.transData.transform((0, maxy)))[1]
        plot.set_title(mtitle,pad=pad)
    elif title:
        plot.set_title(mtitle,fontsize=fontsize)
    
## qq plot #############################################
    # ax2

    p_toplot = sumstats["scaled_P"]
    
    # sort x,y for qq plot
    minit=1/len(p_toplot)
    observed = p_toplot.sort_values(ascending=False)
    expected = -np.log10(np.linspace(minit,1,len(observed)))
    
    
    ax2.scatter(expected,observed,s=8,color=colors[0])
    ax2.plot([0,-np.log10(minit)],[0,-np.log10(minit)],linestyle="--",color=sig_line_color)
    
    ax2.set_xlabel("Expected $-log_{10}(P)$",fontsize=fontsize)
    ax2.set_ylabel("Observed $-log_{10}(P)$",fontsize=fontsize)
    ax2.spines["top"].set_visible(False)
    ax2.spines["right"].set_visible(False)
    ax2.spines["left"].set_visible(True)
    if gc:
        observedMedianChi2 = sp.stats.chi2.isf( np.median(np.power(10,-p_toplot)) ,1)
        expectedMedianChi2 = sp.stats.chi2.ppf(0.5,1)
        lambdagc=observedMedianChi2/expectedMedianChi2
        ax2.text(0.05, 0.95,"$\\lambda_{GC}$ = "+"{:.4f}".format(lambdagc),
                 horizontalalignment='left',
                 verticalalignment='top',
                 transform=ax2.transAxes,
                 fontsize=fontsize)
    if cut:
        qcutline=ax2.axhline(y=cut, linewidth = 2,linestyle="--",color=cut_line_color,zorder=1)
        ax2.set_yticks([x for x in range(0,cut+1,2)]+[(maxticker-cut)/cutfactor + cut])
        ax2.set_yticklabels([x for x in range(0,cut+1,2)]+[maxticker])

    if qtitle:
        ax2.set_title(qtitle,fontsize=fontsize,pad=10)

    if verbose: print("  - Created QQ plot successfully!")
################################################################
    ## return figure
    if save:
        if verbose: print("Saving plot:")
        if save==True:
            fig.savefig("./manhattan_qq_plot.png",bbox_inches="tight",**saveargs)
            print("  - Saved to "+ "./manhattan_qq_plot.png" + " successfully!" )
        else:
            fig.savefig(save,bbox_inches="tight",**saveargs)
            print("  - Saved to "+ save + " successfully!" )
    if title and anno and len(to_annotate)>0:
        fig.suptitle(title ,x=0.5, y=1.2)
    else:
        fig.suptitle(title ,x=0.5,y=1)
###################################################################    
    return fig


import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import numpy as np
from gwaslab import getsig

def mplot(insumstats,
          chrom,
          pos,
          p,
          scaled=False,
          cut=0,
          cutfactor=10,
          cut_line_color="#ebebeb",
          windowsizekb=500,
          anno=None,
          sig_level=5e-8,
          sig_line_color="grey",
          suggestive_sig_level=5e-6,
          title =None,
          mtitle=None,
          figsize =(15,5),
          fontsize = 10,
          colors = ["#000042", "#7878BA"],
          verbose=True,
          repel_force=0.03,
          save=None,
          saveargs={"dpi":300,"facecolor":"white"}
          ):
    
# printing meta info ###############################
    if verbose: print("Basic settings:")
    if verbose: print("  - Genome-wide significance level is set to "+str(sig_level)+" ...")
    if verbose: print("  - Raw input contains "+str(len(insumstats))+" variants...")
    if verbose: print("Start conversion and QC:")
    
# read sumstats ###############################
    sumstats=pd.DataFrame()
    sumstats["P-value_association"]=insumstats[p].astype(np.float64)
    sumstats["CHROM"]=insumstats[chrom].astype(int)
    sumstats["POS"]=insumstats[pos].astype(int)
    if anno and anno!=True:
            sumstats["Annotation"]=insumstats[anno].astype(object)
            
# P value conversion ###############################  
    if scaled:
        if verbose:print("  - P values are already converted to -log10(P)!")
        sumstats["scaled_P"] = sumstats["P-value_association"]
    else:
        if verbose:print("  - P values are being converted to -log10(P)...")
        #sanity check
        outside_01_p_value_number=len(sumstats.loc[(sumstats["P-value_association"]>1)|(sumstats["P-value_association"]<=0),:])
        if verbose:print("  - Sanity check: "+ str(outside_01_p_value_number) +" variants with P value outside of (0,1] will be removed...")
        sumstats = sumstats.loc[(sumstats["P-value_association"]<=1)&(sumstats["P-value_association"]>0),:]
        sumstats["scaled_P"] = -np.log10(sumstats["P-value_association"])

    if verbose: print("  - Sanity check: "+str(len(sumstats[sumstats["scaled_P"].isin([np.nan, np.inf, -np.inf, float('inf'), None])])) + " na/inf/-inf variants will be removed..." )
    sumstats = sumstats[~sumstats["scaled_P"].isin([np.nan, np.inf, -np.inf, float('inf'), None])]

    # shrink at a certain value #############################################
    maxy = sumstats["scaled_P"].max()
    if verbose: print("  - Maximum -log10(P) values are "+str(maxy) +" .")
    if cut:
        if verbose: print("  - Minus log10(P) values above " + str(cut)+" will be shrunk with a shrinkage factor of " + str(cutfactor)+"...")
        maxticker=int(np.round(sumstats["scaled_P"].max()))
        sumstats.loc[sumstats["scaled_P"]>cut,"scaled_P"] = (sumstats.loc[sumstats["scaled_P"]>cut,"scaled_P"]-cut)/cutfactor +  cut
        maxy = (maxticker-cut)/cutfactor + cut

# create index ##########################

    if verbose:print("Plotting "+str(len(sumstats))+" variants:")
    #sort & add id
    sumstats = sumstats.sort_values(["CHROM","POS"])
    sumstats["id"]=range(len(sumstats))
    sumstats=sumstats.set_index("id")

    #create a position dictionary
    posdic=sumstats.groupby("CHROM")["POS"].max()
    posdiccul=dict(posdic)
    posdiccul[0]=0
    for i in range(2,sumstats["CHROM"].nunique()+1):
        posdiccul[i]=posdiccul[i-1]+posdiccul[i]

    #convert base pair postion to x axis position
    sumstats["add"]=sumstats["CHROM"].apply(lambda x : posdiccul[int(x)-1])
    sumstats["i"]=sumstats["POS"]+sumstats["add"]

    #for plot
    chrom_df=sumstats.groupby('CHROM')['i'].median()
    chrom_df=chrom_df+((chrom_df.index.astype(int))-1)*len(sumstats)*0.02

    sumstats["i"]=sumstats["i"]+((sumstats["CHROM"].astype(int))-1)*len(sumstats)*0.01

# assign marker size ##############################################
    sumstats["s"]=1
    sumstats.loc[sumstats["scaled_P"]>-np.log10(5e-4),"s"]=2
    sumstats.loc[sumstats["scaled_P"]>-np.log10(suggestive_sig_level),"s"]=3
    sumstats.loc[sumstats["scaled_P"]>-np.log10(sig_level),"s"]=10

# Plotting ########################################################  
    # manhattanplot : ax1
    fig,  ax1 = plt.subplots(1,1,figsize=figsize)
    plot = sns.scatterplot(data=sumstats, x='i', y='scaled_P',
                       hue='CHROM',
                       palette=sns.color_palette(colors,n_colors=sumstats["CHROM"].nunique()),
                       legend=None,
                       size="s",
                       sizes=(10,40),
                       linewidth=0,
                       zorder=2,ax=ax1)   
        
    plot.set_xlabel('CHROM'); 
    plot.set_xticks(chrom_df);
    plot.set_xticklabels(chrom_df.index)
    
    sigline = plot.axhline(y=-np.log10(sig_level), linewidth = 2,linestyle="--",color=sig_line_color,zorder=1)
    if cut:
        cutline=plot.axhline(y=cut, linewidth = 2,linestyle="--",color=cut_line_color,zorder=1)
        plot.set_yticks([x for x in range(0,cut+1,2)]+[(maxticker-cut)/cutfactor + cut])
        plot.set_yticklabels([x for x in range(0,cut+1,2)]+[maxticker])

# Annotation #######################################################
    #!!!!!!!!!!!!!!! scale problem
    if anno and anno!=True:
        to_annotate=getsig(sumstats.loc[sumstats["scaled_P"]>-np.log10(sig_level)],
                           "Annotation",
                           'CHROM',
                           "POS",
                           "P-value_association",
                           sig_level=sig_level,
                           windowsizekb=windowsizekb,
                           verbose=False)
    else:
        to_annotate=getsig(sumstats.loc[sumstats["scaled_P"]>-np.log10(sig_level)],
                           "i",
                           'CHROM',
                           "POS",
                           "P-value_association",
                           windowsizekb=windowsizekb,
                           verbose=False,
                           sig_level=sig_level)
        
    if verbose:print("  - Found "+str(len(to_annotate))+" significant variants with a sliding window size of "+str(windowsizekb)+" kb...")

# Annotation #######################################################
    
    if anno and len(to_annotate)>0:
        #initiate a list for text and a starting position
        text = []
        last_pos=0
        
        for rowi,row in to_annotate.iterrows():
            # avoid text overlapping
            if row["i"]>last_pos+repel_force*sumstats["i"].max():
                last_pos=row["i"]
            else:
                last_pos+=repel_force*sumstats["i"].max()
            # data to pixels
            armB_length_in_point=plot.transData.transform((0, 0.95*maxy))[1]-plot.transData.transform((0, row["scaled_P"]+1))[1]
            
            #  
            armB_length_in_point= armB_length_in_point if armB_length_in_point>0 else plot.transData.transform((0, maxy+2))[1]-plot.transData.transform((0,  row["scaled_P"]+1))[1] 
            
            if anno==True:
                annotation_text=str(int(row["CHROM"])) +":"+ str(int(row["POS"]))
                annotation_col="CHR:POS"
            elif anno:
                annotation_text=row["Annotation"]
                annotation_col=anno
                
            text.append(plot.annotate(annotation_text,
                                         xy=(row["i"],row["scaled_P"]+1),
                                         xytext=(last_pos,1.15*maxy),rotation=40,
                                         ha="left",va="bottom",
                                         fontsize=fontsize,
                                         arrowprops=dict(arrowstyle="-|>",relpos=(0,0),color="#ebebeb",
                                         connectionstyle="arc,angleA=0,armA=0,angleB=90,armB="+str(armB_length_in_point)+",rad=0")
                                        )
                       )
        
        if verbose: print("  - Annotating using column "+annotation_col+"...")
    else:
        if verbose: print("  - Skip annotating")

    plot.set_ylabel("$-log_{10}(P)$",fontsize=fontsize)
    plot.set_xlabel("Chromosomes",fontsize=fontsize)
    plot.spines["top"].set_visible(False)
    plot.spines["right"].set_visible(False)
    plot.spines["left"].set_visible(True)
    
    if verbose: print("  - Created Manhattan plot successfully!")
    if mtitle and anno and len(to_annotate)>0: 
        pad=(plot.transData.transform((0, 1.35*maxy))[1]-plot.transData.transform((0, maxy)))[1]
        plot.set_title(mtitle,pad=pad)
    elif title:
        plot.set_title(mtitle,fontsize=fontsize)
    
    ## return figure
    if save:
        if verbose: print("Saving plot:")
        if save==True:
            fig.savefig("./manhattanplot.png",bbox_inches="tight",**saveargs)
            print("  - Saved to "+ "./manhattanplot.png" + " successfully!" )
        else:
            fig.savefig(save,bbox_inches="tight",**saveargs)
            print("  - Saved to "+ save + " successfully!" )
    return plot


from gwaslab.getsig import getsig
from gwaslab.manhattan import mplot
from gwaslab.qqplot import qqplot
from gwaslab.mqqplot import mqqplot
from gwaslab.qqplot import gc
from gwaslab.get_hapmap3 import ldsc
from gwaslab.h2_conversion import h2_obs_to_liab
from gwaslab.compare_effect import compare_effect
from gwaslab.read_ldsc import read_ldsc



import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import numpy as np
import scipy as sp
def getsig(insumstats,id,chrom,pos,p,windowsizekb=500,verbose=True,sig_level=5e-8):
    
    if verbose: print("  - Processing "+str(len(insumstats))+" variants...")
    sumstats=insumstats.loc[~insumstats[id].isna(),:]
    sumstats_sig = sumstats.loc[sumstats[p]<sig_level,:]
    if verbose:print("  - Found "+str(len(sumstats_sig))+" significant variants with a sliding window size of "+str(windowsizekb)+" kb...")
    sumstats_sig = sumstats_sig.sort_values([chrom,pos])
    
    if len(sumstats_sig)==0:
        if verbose:print("  - No lead snps at given significance threshold!")
        return None
    
    sig_index_list=[]
    current_sig_index = False
    current_sig_p = 1
    current_sig_pos = 0
    current_sig_chr = 0
    for line_number,(index, row) in enumerate(sumstats_sig.iterrows()):
        #when finished one chr 
        if row[chrom]>current_sig_chr:
            if current_sig_index:sig_index_list.append(current_sig_index)
            current_sig_chr=row[chrom]
            current_sig_pos=row[pos]
            current_sig_p=row[p]
            current_sig_index=row[id]
            continue
        
        #next loci
        if row[pos]>current_sig_pos + windowsizekb*1000:
            sig_index_list.append(current_sig_index)
            current_sig_pos=row[pos]
            current_sig_p=row[p]
            current_sig_index=row[id]
            continue
        # update current pos and p
        if row[p]<current_sig_p:
            current_sig_pos=row[pos]
            current_sig_p=row[p]
            current_sig_index=row[id]
        else:
            current_sig_pos=row[pos]
        #when last line
        if  line_number == len(sumstats_sig)-1:
            sig_index_list.append(current_sig_index)
            continue
    if verbose:print("  - Identified "+str(len(sig_index_list))+" lead variants!")
    return sumstats_sig.loc[sumstats_sig[id].isin(sig_index_list),:]


import sys
import gwaslab as gl
import os, psutil
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from adjustText import adjust_text

def compare_effect(path1,cols_name_list_1,
                   path2,cols_name_list_2,
                   label1="Sumstats_1",
                   label2="Sumstats_2",
                   label3="Both",
                   snplist=None,
                   anno=False,
                   verbose=False,
                   sig_level=5e-8,
                   scatterargs={"s":30},
                   errargs={"ecolor":"grey","elinewidth":1}):
    
    #[snpid,chr,pos,p,ea,nea,effect,se]

    
    #load sumstats1
    if verbose: print("Loading "+label1+"...")
    sumstats = pd.read_table(path1,sep="\s+",usecols=[cols_name_list_1[0],
                                                       cols_name_list_1[1],
                                                       cols_name_list_1[2],
                                                       cols_name_list_1[3]])
    sumstats.rename(columns={
                        cols_name_list_1[0]:"SNPID",
                        cols_name_list_1[1]:"CHR",
                        cols_name_list_1[2]:"POS",
                        cols_name_list_1[3]:"P"
                              },
                     inplace=True)
    
    #extact SNPs for comparison 
    if verbose: print("Extract top snps from "+label1+"...")
    
    #if a snplist is provided, use the snp list
    if snplist: 
        if verbose: print("Extract snps in the given list from "+label1+"...")
        sig_list_1 = sumstats.loc[sumstats["SNPID"].isin(snplist),:]
    else:
        # otherwise use the sutomatically detected lead SNPs
        sig_list_1 = gl.getsig(sumstats,"SNPID","CHR","POS","P",
                               verbose=verbose,sig_level=sig_level)

    # do the same for sumstats 2
    if verbose: print("Loading "+label2+"...")
    sumstats = pd.read_table(path2,sep="\s+",usecols=[cols_name_list_2[0],
                                                       cols_name_list_2[1],
                                                       cols_name_list_2[2],
                                                       cols_name_list_2[3]])
    sumstats.rename(columns={
                        cols_name_list_2[0]:"SNPID",
                        cols_name_list_2[1]:"CHR",
                        cols_name_list_2[2]:"POS",
                        cols_name_list_2[3]:"P"
                              },
                     inplace=True)   
    
    if verbose: print("Extract top snps from "+label2+"...")
    if snplist: 
        if verbose: print("Extract snps in the given list from "+label2+"...")
        sig_list_2 = sumstats.loc[sumstats["SNPID"].isin(snplist),:]
    else: 
        sig_list_2 = gl.getsig(sumstats,"SNPID","CHR","POS","P",
                                 verbose=verbose,sig_level=sig_level)
    
    
    #Merge two list using SNPID
    ##############################################################################
    if verbose: print("Merging snps from "+label1+" and "+label2+"...")
    sig_list_merged = pd.merge(sig_list_1,sig_list_2,left_on="SNPID",right_on="SNPID",how="outer",suffixes=('_1', '_2'))
    sig_list_merged["indicator"] = 0
    sig_list_merged.loc[sig_list_merged["P_1"]<sig_level,"indicator"]+=1
    sig_list_merged.loc[sig_list_merged["P_2"]<sig_level,"indicator"]+=2
    ###############################################################################
    
    #### merging sumstats1
    if verbose: print("Extract the EFFECT_ALLELE, NON_EFFECT_ALLELE, EFFECT_SIZE,and SE of selected snps from "+label1+"...")
    sumstats = pd.read_table(path1,sep="\s+",
                              usecols=[cols_name_list_1[0],
                                       cols_name_list_1[4],
                                       cols_name_list_1[5],
                                       cols_name_list_1[6],
                                       cols_name_list_1[7]])
    sumstats.rename(columns={
                        cols_name_list_1[0]:"SNPID",
                        cols_name_list_1[4]:"EA_1",
                        cols_name_list_1[5]:"NEA_1",
                        cols_name_list_1[6]:"EFFECT_1",
                        cols_name_list_1[7]:"SE_1",
                              },
                     inplace=True)
    
    if verbose: print("Merging "+label1+" effect information...")
    sig_list_merged = pd.merge(sig_list_merged,sumstats,
                               left_on="SNPID",right_on="SNPID",
                               how="left")
    
    #### merging sumstats2
    if verbose: print("Extract the  EFFECT_ALLELE, NON_EFFECT_ALLELE, EFFECT_SIZE,and SE of selected snps from "+label2+"...")
    sumstats = pd.read_table(path2,sep="\s+",
                              usecols=[cols_name_list_2[0],
                                       cols_name_list_2[4],
                                       cols_name_list_2[5],
                                       cols_name_list_2[6],
                                       cols_name_list_2[7]])
    
    sumstats.rename(columns={
                        cols_name_list_2[0]:"SNPID",
                        cols_name_list_2[4]:"EA_2",
                        cols_name_list_2[5]:"NEA_2",
                        cols_name_list_2[6]:"EFFECT_2",
                        cols_name_list_2[7]:"SE_2"
                              },
                     inplace=True)
    if verbose: print("Merging "+label2+" effect information...")
    sig_list_merged = pd.merge(sig_list_merged,sumstats,
                               left_on="SNPID",right_on="SNPID",
                               how="left")
    
    sig_list_merged.set_index("SNPID",inplace=True)
    
    ###update sumstats1
    sumstats = pd.read_table(path1,sep="\s+",usecols=[cols_name_list_1[0],cols_name_list_1[3]])
    sumstats.rename(columns={
                        cols_name_list_1[0]:"SNPID",
                        cols_name_list_1[3]:"P_1"
                              },
                     inplace=True)
    sumstats.set_index("SNPID",inplace=True)
    sig_list_merged.update(sumstats)
    
    ###update sumstats2
    sumstats = pd.read_table(path2,sep="\s+",usecols=[cols_name_list_2[0],cols_name_list_2[3]])
    sumstats.rename(columns={
                        cols_name_list_2[0]:"SNPID",
                        cols_name_list_2[3]:"P_2"
                              },
                     inplace=True)
    sumstats.set_index("SNPID",inplace=True)
    sig_list_merged.update(sumstats)
    
    ####
    sig_list_merged["CHR"]=np.max(sig_list_merged[["CHR_1","CHR_2"]], axis=1).astype(int)
    sig_list_merged["POS"]=np.max(sig_list_merged[["POS_1","POS_2"]], axis=1).astype(int)
    sig_list_merged.drop(labels=['CHR_1', 'CHR_2','POS_1', 'POS_2'], axis=1,inplace=True)
    
    #### align allele effect with sumstats 1
    sig_list_merged["EA_2_aligned"]=sig_list_merged["EA_2"]
    sig_list_merged["NEA_2_aligned"]=sig_list_merged["NEA_2"]
    sig_list_merged["EFFECT_2_aligned"]=sig_list_merged["EFFECT_2"]

    sig_list_merged.loc[sig_list_merged["EA_1"]!=sig_list_merged["EA_2"],"EA_2_aligned"]= sig_list_merged.loc[sig_list_merged["EA_1"]!=sig_list_merged["EA_2"],"NEA_2"]
    sig_list_merged.loc[sig_list_merged["EA_1"]!=sig_list_merged["EA_2"],"NEA_2_aligned"]= sig_list_merged.loc[sig_list_merged["EA_1"]!=sig_list_merged["EA_2"],"EA_2"]
    sig_list_merged.loc[sig_list_merged["EA_1"]!=sig_list_merged["EA_2"],"EFFECT_2_aligned"]= -sig_list_merged.loc[sig_list_merged["EA_1"]!=sig_list_merged["EA_2"],"EFFECT_2"]

    sum1only = sig_list_merged.loc[sig_list_merged["indicator"]==1,:].dropna(axis=0)
    sum2only = sig_list_merged.loc[sig_list_merged["indicator"]==2,:].dropna(axis=0)
    both = sig_list_merged.loc[sig_list_merged["indicator"]==3,:].dropna(axis=0)
    
    if verbose: print("Identified "+str(len(sum1only)) + " variants which are only significant in " + label1+".")
    if verbose: print("Identified "+str(len(sum2only)) + " variants which are only significant in " + label2+".")
    if verbose: print("Identified "+str(len(both)) + " variants which are significant in " + label3 + ".")
    
    ##plot########################################################################################
    if verbose: print("Plotting the scatter plot for effect size comparison...")
    plt.style.use("default")
    fig,ax = plt.subplots(figsize=(8,8)) 
    
    if len(sum1only)>0:
        ax.errorbar(sum1only["EFFECT_1"],sum1only["EFFECT_2_aligned"], xerr=sum1only["SE_1"],yerr=sum1only["SE_2"],
                    linewidth=0,zorder=1,**errargs)
        ax.scatter(sum1only["EFFECT_1"],sum1only["EFFECT_2_aligned"],label=label1,zorder=2,**scatterargs)
        
    if len(sum2only)>0:
        ax.errorbar(sum2only["EFFECT_1"],sum2only["EFFECT_2_aligned"], xerr=sum2only["SE_1"],yerr=sum2only["SE_2"],
                    linewidth=0,zorder=1,**errargs)
        ax.scatter(sum2only["EFFECT_1"],sum2only["EFFECT_2_aligned"],label=label2,zorder=2,**scatterargs)

    if len(both)>0:
        ax.errorbar(both["EFFECT_1"],both["EFFECT_2_aligned"], xerr=both["SE_1"],yerr=both["SE_2"],
                    linewidth=0,zorder=1,**errargs)
        ax.scatter(both["EFFECT_1"],both["EFFECT_2_aligned"],label=label3,zorder=2,**scatterargs)
    
    ## annotation
    if anno==True:
        sig_list_toanno = sig_list_merged.dropna(axis=0)
        texts=[]
        for index, row in sig_list_toanno.iterrows():
            texts.append(plt.text(row["EFFECT_1"], row["EFFECT_2_aligned"],index, ha='center', va='center')) 
        adjust_text(texts, arrowprops=dict(arrowstyle='->', color='grey'))
    elif type(anno) is dict:
        # if input is a dict
        sig_list_toanno = sig_list_toanno.loc[sig_list_toanno.index.isin(list(anno.keys())),:]
        texts=[]
        for index, row in sig_list_toanno.iterrows():
            texts.append(plt.text(row["EFFECT_1"], row["EFFECT_2_aligned"],anno[index], ha='right', va='top')) 
        adjust_text(texts,ha='right', va='top',arrowprops=dict(arrowstyle='->', color='grey'))
        
    ax.set_xlabel("Effect sizes in "+label1)
    ax.set_ylabel("Effect sizes in "+label2)
    
    # plot x=0,y=0, and a 45 degree line
    xl,xh=ax.get_xlim()
    yl,yh=ax.get_ylim()
    ax.axline([min(xl,yl),min(xl,yl)], [max(xh, yh),max(xh, yh)],color='grey', linestyle='--')
    ax.axhline(y=0, color='grey', linestyle='--')
    ax.axvline(x=0, color='grey', linestyle='--')
    ax.legend()
    ##plot finished########################################################################################
    
    return [sig_list_merged, fig]
    #get top variants from sumstats1


import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import numpy as np
import scipy as sp


def qqplot(insumstats,
           p,
           title="QQ plot",
           figsize=(8,8),
           fontsize=15,
           color="#000042",
           s=10,
           gc=True,
           save=None,
           saveargs={"dpi":300,"facecolor":"white"},
           sig_line_color="grey",
           verbose=True,
           scaled=False,
           rtspine=True
          ):
    sumstats=insumstats.loc[~insumstats[p].isna(),p].astype(np.float64)
    if verbose: print(" - Processing "+ str(len(sumstats)) +" non-missing variants...")
    
    if scaled:
        if verbose:print(" - P values are already converted to -log10(P).")
        p_toplot = sumstats
    else:
        if verbose:print(" - P values are being converted to -log10(P)...")
        p_toplot = -np.log10(sumstats)    

    # sort x,y for qq plot
    minit=1/len(p_toplot)
    observed = p_toplot.sort_values(ascending=False)
    expected = -np.log10(np.linspace(minit,1,len(observed)))
    
    fig, ax2 = plt.subplots(1, 1,figsize=figsize)
    ax2.scatter(expected,observed,s=s,color=color)
    ax2.plot([0,-np.log10(minit)],[0,-np.log10(minit)],linestyle="--",color=sig_line_color)
    ax2.set_xlabel("Expected $-log_{10}(P)$",fontsize=fontsize)
    ax2.set_ylabel("Observed $-log_{10}(P)$",fontsize=fontsize)
    if not rtspine:
        ax2.spines["top"].set_visible(False)
        ax2.spines["right"].set_visible(False)
    ax2.spines["left"].set_visible(True)
    
    if gc:
        observedMedianChi2 = sp.stats.chi2.isf( np.median(np.power(10,-p_toplot)) ,1)
        expectedMedianChi2 = sp.stats.chi2.ppf(0.5,1)
        lambdagc=observedMedianChi2/expectedMedianChi2
        ax2.text(0.05, 0.95,"$\\lambda_{GC}$ = "+"{:.4f}".format(lambdagc),
                 horizontalalignment='left',
                 verticalalignment='top',
                 transform=ax2.transAxes,
                 fontsize=fontsize)

    if verbose: print(" - Created QQ plot successfully!")
    if title:
        plt.title(title,fontsize=fontsize,pad=10)
    if save:
        if verbose: print("Saving plot:")
        if save==True:
            plt.savefig("./qqplot.png",bbox_inches="tight",**saveargs)
            if verbose: print(" - Saved to "+ "./qqplot.png" + " successfully!" )
        else:    
            plt.savefig(save,bbox_inches="tight",**saveargs)
            if verbose: print(" - Saved to "+ save + " successfully!" )
    return fig
    
def maf_qqplot():
    return 0

def gc(insumstats,mode="p",level=0.5):
    indata=insumstats.dropna()
    if mode=="p" or mode=="P":
        observedMedianChi2 = sp.stats.chi2.isf(np.median(indata),1)
        expectedMedianChi2 = sp.stats.chi2.ppf(level,1)
        lambdagc=observedMedianChi2/expectedMedianChi2
        print("(p mode) Lambda GC at "+ str(level)+ " is"," ","{:.5f}".format(lambdagc))
    elif mode=="z" or mode=="Z":
        observedMedianChi2 = np.median((indata)**2)
        expectedMedianChi2 = sp.stats.chi2.ppf(level,1)
        lambdagc=observedMedianChi2/expectedMedianChi2
        print("(z mode) Lambda GC at "+ str(level)+ " is"," ","{:.5f}".format(lambdagc))
    elif mode=="chi2" or mode=="chi2":
        observedMedianChi2 = np.median(indata)
        expectedMedianChi2 = sp.stats.chi2.ppf(level,1)
        lambdagc=observedMedianChi2/expectedMedianChi2
        print("(chi2 mode) Lambda GC at "+ str(level)+ " is"," ","{:.5f}".format(lambdagc))
    else:
        return None
    return lambdagc

    